# -*- python -*-

from omniidl import idlvisitor, idlast;
from omniidl import output;
import os;
import conversionutils;
import simplecxx;

class Walker(idlvisitor.AstVisitor):
    """Walks over the AST once and writes the CCI header or CCI as it goes.
    """


    
    def visitAST(self, node):
        """Visit all the declarations in an AST"""

        self.beenIncluded = {}
        self.masterFile = node.file()
        self.masterGuard = '';
        self.scope = ["CCI"]
        self.scopeEntryDepth = 0
        for i in node.filebase:
            if (i >= 'a' and i <= 'z') or (i >= 'A' and i <= 'Z'):
                self.masterGuard = self.masterGuard + i
        self.cci.out("/* This file is automatically generated from " +\
                     node.filename + """
 * DO NOT EDIT DIRECTLY OR CHANGES WILL BE LOST!
 */""")
        if not self.doing_header:
            self.cci.out("""
#include <omniORB4/CORBA.h>
#include <inttypes.h>
#include <strings.h>
#include <wchar.h>
#include <exception>""")
            self.cci.out('#include "CCI' + node.filebase + '.hxx"')
        else:
            self.cci.out('#include "Iface' + node.filebase + '.hxx"')
            self.cci.out('#include "' + node.filebase + '.hh"')
        for n in node.declarations():
            if n.file() == self.masterFile:
                n.accept(self)
            elif self.doing_header:
                self.considerIncluding(n.file())
        if self.doing_header:
            self.escapeScopes()

    def escapeScopes(self):
        for i in range(0, self.scopeEntryDepth):
            self.cci.dec_indent()
            self.cci.out('};')
        self.scopeEntryDepth = 0

    def writeScopes(self):
        for i in range(self.scopeEntryDepth, len(self.scope)):
            self.cci.out('namespace ' + self.scope[i])
            self.cci.out('{')
            self.cci.inc_indent()
        self.scopeEntryDepth = len(self.scope)

    def enterScope(self, node):
        self.scope.append(node.simplename)

    def leaveScope(self):
        self.scope = self.scope[:-1]
        if self.scopeEntryDepth > len(self.scope):
            self.scopeEntryDepth = len(self.scope)
            self.cci.dec_indent()
            self.cci.out('};')

    def considerIncluding(self, name):
        if (self.beenIncluded.has_key(name)):
            return
        self.beenIncluded[name] = 1;
        self.escapeScopes()
        basename,ext = os.path.splitext(name)
        self.cci.out('#include "CCI' + basename  + '.hxx"')
        self.cci.inModule = False

    def visitModule(self, node):
        """Visit all the definitions in a module."""
        self.enterScope(node)
        # self.cci.out('namespace ' + node.simplename)
        # self.cci.out('{')
        # self.cci.inc_indent()

        for n in node.definitions():
            if n.file() == self.masterFile:
                if self.doing_header:
                    self.writeScopes()
                n.accept(self)
            elif self.doing_header:
                self.considerIncluding(n.file())
        # self.cci.dec_indent()
        # self.cci.out('}')
        self.leaveScope()

    def processBase(self, active, base):
        # Interface active has base(or active == base). Called only once
        # per active per base. We only care about callables here.
        for c in base.callables():
            if isinstance(c, idlast.Operation):
                self.processBaseOperation(active, base, c)
            else:
                self.processBaseAttribute(active, base, c)

    def writeAddRef(self):
        self.cci.out('{')
        self.cci.inc_indent()
        self.cci.out('_refcount++;')
        self.cci.dec_indent()
        self.cci.out('}')

    def writeReleaseRef(self):
        self.cci.out('{')
        self.cci.inc_indent()
        self.cci.out('_refcount--;')
        self.cci.out('if (_refcount == 0)')
        self.cci.out('{')
        self.cci.inc_indent()
        # Send the remove_ref upstream...
        self.cci.out('try')
        self.cci.out('{')
        self.cci.inc_indent()        
        self.cci.out('_objref->release_ref();')
        self.cci.dec_indent()
        self.cci.out('}')
        self.cci.out('catch (CORBA::Exception& e)')
        self.cci.out('{')
        self.cci.out('}')        
        self.cci.out('delete this;')
        self.cci.dec_indent()
        self.cci.out('}')
        self.cci.dec_indent()
        self.cci.out('}')

    def processBaseOperation(self, active, base, op):
        psemi = ''
        pfq = 'CCI::' + active.corbacxxscoped + '::'
        if self.doing_header:
            psemi = ';'
            pfq = ''
        # Firstly, get the return type...
        rtype = simplecxx.typeToSimpleCXX(op.returnType())
        # Next, build up all the parameters...
        parstr = ''
        needcomma = False
        for p in op.parameters():
            if needcomma:
                parstr = parstr + ', '
            else:
                needcomma = True
            if p.is_out():
                extrapointer = 1
            else:
                extrapointer = 0
            parstr = parstr + simplecxx.typeToSimpleCXX(p.paramType(), \
                                                        extrapointer, not p.is_out()) +\
                     ' ' + p.simplename
            if simplecxx.doesTypeNeedLength(p.paramType()):
                parstr = parstr + ', uint32_t'
                if p.paramType().is_out():
                    parstr = parstr + '*'
                parstr = parstr + ' _length_' + p.simplename
        
        if simplecxx.doesTypeNeedLength(op.returnType()):
            if needcomma:
                parstr = parstr + ', '
            parstr = parstr + 'uint32_t* _length_' + p.simplename
        
        self.cci.out(rtype + ' ' + pfq + op.simplename + '(' + parstr +\
                     ') throw(std::exception&)' + psemi)
        if self.doing_header:
            return
        self.cci.ci_count = 0
        if op.simplename == 'add_ref':
            self.writeAddRef()
        elif op.simplename == 'release_ref':
            self.writeReleaseRef()
        else:
            self.cci.out('{')
            self.cci.inc_indent()
            # All in parameters get converted to CORBA parameters
            for p in op.parameters():
                self.declareCORBAStorage(p.paramType(), '_corba_' + p.simplename)
                # If it isn't an in parameter, leave it for now...
                if not p.is_in():
                    continue
                if simplecxx.doesTypeNeedLength(p.paramType()):
                    self.simpleSequenceToCORBA(p.paramType(), p.simplename, \
                                               '_length_' + p.simplename, \
                                               '_corba_' + p.simplename)
                else:
                    self.simpleValueToCORBA(p.paramType(), p.simplename, \
                                            '_corba_' + p.simplename)
            # Next, make the call...
            self.cci.out('try')
            self.cci.out('{')
            self.cci.inc_indent()
            self.cci.dec_indent()
            self.cci.out('}')
            self.cci.out('catch (CORBA::Exception& e)')
            self.cci.out('{')
            self.cci.out('}')

    def processBaseAttribute(self, active, base, at):
        psemi = ''
        pfq = 'CCI::' + active.corbacxxscoped + '::'
        if self.doing_header:
            psemi = ';'
            pfq = ''
        typename = simplecxx.typeToSimpleCXX(at.attrType())
        typenameC = simplecxx.typeToSimpleCXX(at.attrType(), is_const=True)
        if simplecxx.doesTypeNeedLength(at.attrType()):
            extra_getter_params = 'uint32_t* _length_attr'
            extra_setter_params = ', uint32_t _length_attr'
        else:
            extra_getter_params = ''
            extra_setter_params = ''
        for n in at.declarators():
            self.cci.out(typename + ' ' + pfq + n.simplename + '(' +
                         extra_getter_params +
                         ') throw(std::exception&)' + psemi)
            if not self.doing_header:
                self.cci.out('{')
                self.cci.inc_indent()
                self.cci.out('try')
                self.cci.out('{')
                self.cci.inc_indent()
                # 1) Declare storage...
                self.declareCORBAStorage(at.attrType(), '_corba_value')
                # 2) Call the getter and assign...
                self.cci.out('_corba_value = _objref->' + n.simplename + '();')
                self.cci.ci_count = 0
                if simplecxx.doesTypeNeedLength(at.attrType()):
                    self.CORBASequenceToSimple(at.attrType(), '_corba_value', None, '(*_length_attr)')
                else:
                    self.CORBAValueToSimple(at.attrType(), '_corba_value', None)
                self.cci.dec_indent()
                self.cci.out('}')
                self.cci.out('catch (CORBA::Exception& cce)')
                self.cci.out('{')
                self.cci.inc_indent()
                self.cci.out('throw std::exception(/*"A CORBA error occurred"*/);')
                self.cci.dec_indent()
                self.cci.out('}')
                self.cci.dec_indent()
                self.cci.out('}')
            if not at.readonly():
                self.cci.out('void ' + pfq + n.simplename + '(' + typenameC +
                             ' attr' + extra_setter_params +
                             ') throw(std::exception&)' + psemi)
                if not self.doing_header:
                    self.cci.out('{')
                    self.cci.inc_indent()
                    self.cci.out('try')
                    self.cci.out('{')
                    self.cci.inc_indent()
                    self.cci.ci_count = 0
                    self.declareCORBAStorage(at.attrType(), '_corba_value')
                    if simplecxx.doesTypeNeedLength(at.attrType()):
                        self.simpleSequenceToCORBA(at.attrType(), 'attr',
                                                   '_length_attr', '_corba_value')
                    else:
                        self.simpleValueToCORBA(at.attrType(), 'attr',
                                                '_corba_value')
                    # Finally, call the setter...
                    self.cci.out('_objref->' + n.simplename + '(_corba_value);')
                    self.cci.dec_indent()
                    self.cci.out('}')
                    self.cci.out('catch (CORBA::Exception& cce)')
                    self.cci.out('{')
                    self.cci.inc_indent()
                    self.cci.out('throw std::exception(/*"A CORBA error occurred"*/);')
                    self.cci.dec_indent()
                    self.cci.out('}')
                    self.cci.dec_indent()
                    self.cci.out('}')

    def declareCORBAStorage(self, type, name):
        # We need to get a storage string...
        self.cci.out(conversionutils.getCORBAVarType(type) + ' ' + name + ';')

    def simpleSequenceToCORBA(self, type, sarray, slength, cname):
        conversionutils.writeSimpleSequenceToCORBA(self.cci, type, sarray, slength, cname)

    def simpleValueToCORBA(self, type, sname, cname):
        conversionutils.writeSimpleToCORBA(self.cci, type, sname, cname)

    def CORBASequenceToSimple(self, type, cname, sarray, slength):
        conversionutils.writeCORBASequenceToSimple(self.cci, type, cname, sarray, slength)

    def CORBAValueToSimple(self, type, cname, sname):
        conversionutils.writeCORBAValueToSimple(self.cci, type, cname, sname)

    def visitInterface(self, node):
        if self.doing_header:
            self.cci.out('class ' + node.simplename)
            self.cci.out('  : public ' + node.simplecxxscoped)
            self.cci.out('{')
            self.cci.out('private:')
            self.cci.inc_indent()
            self.cci.out('uint32_t _refcount;')
            self.cci.out('::' + node.corbacxxscoped + '_var _objref;')
            self.cci.out('::PortableServer::POA_var _poa;')
            self.cci.dec_indent()
            self.cci.out('public:')
            self.cci.inc_indent()
            self.cci.out(node.simplename + '(::' + node.corbacxxscoped + \
                         '_ptr _aobjref);');
            self.cci.out('virtual ~' + node.simplename + '(){}');
        else:
            self.cci.out('CCI::' + node.corbacxxscoped + '::' +\
                         node.simplename + '(::' +\
                         node.corbacxxscoped + '_ptr _aobjref)')
            self.cci.out('{')
            self.cci.inc_indent()
            self.cci.out('_objref = _aobjref;')
            self.cci.dec_indent()
            self.cci.out('}')
        # For every interface we need to override every getter/setter/operation
        # from this interface and every interface it inherits from, directly
        # or indirectly. However, we can't do a simple recursive call for this,
        # as we cannot add a common base twice. Instead build a dictionary and
        # a stack as we go.
        stack = [node]
        seen = {node.simplecxxscoped: True}
        self.processBase(node, node)
        while len(stack) != 0:
            current = stack.pop()
            for ifa in current.inherits():
                if not seen.has_key(ifa.simplecxxscoped):
                    seen[ifa.simplecxxscoped] = True
                    stack.append(ifa)
                    self.processBase(node, ifa)

        if self.doing_header:
            self.cci.dec_indent()
            self.cci.out('};')
    
def run(tree):
    w = Walker()
    w.cci = output.Stream(file("CCI" + tree.filebase + ".cxx", "w"), 2)
    w.doing_header = False
    tree.accept(w)
    w.cci = output.Stream(file("CCI" + tree.filebase + ".hxx", "w"), 2)
    w.doing_header = True
    tree.accept(w)
